from io import BytesIO, StringIO
import struct
import numpy as np
from .buffer import bufferize

__all__ = ["HeaderSegment", "DataSegment", "TextSegment", "Fcs"]


class HeaderSegment:
    ranges = {
        "version": (0, 5),
        "text_start": (10, 17),
        "text_end": (18, 25),
        "data_start": (26, 33),
        "data_end": (34, 41),
        "analysis_start": (42, 49),
        "analysis_end": (50, 57),
    }

    def __init__(
        self,
        text_start: int,
        text_end,
        data_start,
        data_end,
        analysis_start=0,
        analysis_end=0,
        ver: str = "FCS3.1",
    ) -> None:
        self.ver = ver
        self.text_start = text_start
        self.text_end = text_end
        self.data_start = data_start
        self.data_end = data_end
        self.analysis_start = analysis_start
        self.analysis_end = analysis_end

    def asdict(self):
        return {
            "text_start": self.text_start,
            "text_end": self.text_end,
            "data_start": self.data_start,
            "data_end": self.data_end,
            "analysis_start": self.analysis_start,
            "analysis_end": self.analysis_end,
        }

    @classmethod
    def from_string(cls, s: bytes):
        s = s.decode("UTF-8")
        v = dict()
        for key, value in cls.ranges.items():
            v[key] = s[value[0] : value[1] + 1].strip()
            if key != "version":
                v[key] = int(v[key])

        ver = v.pop("version")
        return HeaderSegment(
            v["text_start"],
            v["text_end"],
            v["data_start"],
            v["data_end"],
            v["analysis_start"],
            v["analysis_end"],
            ver=ver,
        )

    def _write_locations(self, sp: StringIO, type: str):
        sp.seek(self.ranges[f"{type}_start"][0])
        start_str = str(self.asdict()[f"{type}_start"])
        end_str = str(self.asdict()[f"{type}_end"])

        if len(start_str) > 8 or len(end_str) > 8:
            start_str = "0"
            end_str = "0"

        sp.write((8 - len(start_str)) * " " + start_str)
        sp.seek(self.ranges[f"{type}_end"][0])
        sp.write((8 - len(end_str)) * " " + end_str)

    def to_string(self):
        header = None
        with StringIO() as sp:
            sp.write(self.ver)
            sp.write(" " * 4)
            for t in ["text", "data", "analysis"]:
                self._write_locations(sp, t)
            header = sp.getvalue()

        return header


class TextSegment:
    def __init__(
        self,
        tot: int,
        pnn: list,
        pns: list,
        pnr: list,
        pne: list = None,
        mode: str = "L",
        next_data: int = 0,
        byte_order: str = "1,2,3,4",
        text_start=256,
        delim="/",
        datatype="F",
        data_start=None,
        data_end=None,
        analysis_start=None,
        analysis_end=None,
        stext_start=None,
        stext_end=None,
    ) -> None:
        self.tot = tot
        self.pnn = pnn
        self.pns = pns
        self.pnb = "32"
        self.pne = pne
        self.pnr = pnr
        self.mode = mode
        self.next_data = next_data
        self.byte_order = "1,2,3,4"
        self.text_start = text_start
        self.delim = delim
        self.datatype = datatype
        self.endian = {"1,2,3,4": "<", "4,3,2,1": ">"}[byte_order]
        self._data_start = data_start
        self._data_end = data_end
        self._analysis_start = analysis_start
        self._analysis_end = analysis_end
        self._stext_start = stext_start
        self._stext_end = stext_end
        self.text = self.build()

    def build(self):
        text = {}
        text["BYTEORD"] = self.byte_order
        text["DATATYPE"] = self.datatype
        text["MODE"] = self.mode
        text["NEXTDATA"] = str(self.next_data)
        text["BEGINANALYSIS"] = (
            str(self._analysis_start) if self._analysis_start is not None else "0"
        )
        text["ENDANALYSIS"] = (
            str(self._analysis_end) if self._analysis_end is not None else "0"
        )
        text["BEGINDATA"] = (
            str(self._data_start) if self._data_start is not None else "0"
        )
        text["ENDDATA"] = str(self._data_end) if self._data_end is not None else "0"
        text["BEGINSTEXT"] = (
            str(self._stext_start) if self._stext_start is not None else "0"
        )
        text["ENDSTEXT"] = str(self._stext_end) if self._stext_end is not None else "0"

        text["PAR"] = str(len(self.pnn))
        text["TOT"] = str(self.tot)

        for i, _ in enumerate(self.pnn):
            text["P%dB" % (i + 1)] = self.pnb
            text["P%dE" % (i + 1)] = self.pne
            text["P%dR" % (i + 1)] = str(int(self.pnr[i]))
            text["P%dN" % (i + 1)] = self.pnn[i]
            text["P%dS" % (i + 1)] = self.pns[i]

        if self._data_start is None and self._data_end is None:
            max_text_end = self.cal_max_text_ends(text, self.delim, self.text_start)
            datasize = 4 * len(self.pnn) * self.tot  # 4 is bytes to hold float
            data_start, data_end = self.cal_datapos(max_text_end + 1, datasize)
            text["BEGINDATA"] = data_start
            text["ENDDATA"] = data_end
        return text

    @property
    def data_start(self):
        return int(self.text["BEGINDATA"])

    @property
    def data_end(self):
        return int(self.text["ENDDATA"])

    @property
    def text_end(self):
        return len(self.dict_to_string(self.text, self.delim)) + self.text_start - 1

    @classmethod
    def cal_max_text_ends(cls, text, delim, text_start):
        s = cls.dict_to_string(
            {
                **text,
                "BEGINANALYSIS": "99999999",
                "ENDANALYSIS": "99999999",
                "BEGINDATA": "99999999",
                "ENDDATA": "99999999",
                "BEGINSTEXT": "99999999",
                "ENDSTEXT": "99999999",
            },
            delim,
        )
        return text_start + len(s) - 1

    def to_string(self):
        return self.dict_to_string(self.text, self.delim)

    def update_pns(self, mappings):
        for k, v in mappings.items():
            self.pns[self.pns.index(k)] = v
        self.text = self.build()

    def update_pnn(self, mappings):
        for k, v in mappings.items():
            self.pnn[self.pnn.index(k)] = v
        self.text = self.build()

    @classmethod
    def from_string(self, s):
        delim = s[:1]
        keyvalarray = s[1:].split(delim)
        for num, i in enumerate(keyvalarray):
            try:
                keyvalarray[num] = i.decode("UTF-8")
            except UnicodeDecodeError:
                # print(num, i)
                pass

        vars = {}
        for k, v in zip(keyvalarray[::2], keyvalarray[1::2]):
            vars[k[1:]] = v

        long_channels = list()
        short_channels = list()
        max_values = list()
        for n in range(int(vars.pop("PAR"))):
            short_channels.append(vars.pop("P%dN" % (n + 1)))
            long_channels.append(
                vars.pop("P%dS" % (n + 1)) if "P%dS" % (n + 1) in vars else " "
            )
            max_values.append(int(vars.pop("P%dR" % (n + 1))))

        return TextSegment(
            int(vars["TOT"]),
            short_channels,
            long_channels,
            max_values,
            mode=vars["MODE"],
            next_data=vars["NEXTDATA"],
            byte_order=vars["BYTEORD"],
            delim=delim.decode("UTF-8"),
            datatype=vars["DATATYPE"],
            data_start=vars["BEGINDATA"],
            data_end=vars["ENDDATA"],
            analysis_start=vars["BEGINANALYSIS"],
            analysis_end=vars["ENDANALYSIS"],
            stext_start=vars["BEGINSTEXT"],
            stext_end=vars["ENDSTEXT"],
        )

    @classmethod
    def dict_to_string(cls, dict_, delim):
        # print('convert', dict_)
        s = delim
        for i in dict_:
            # print(type(i), type(dict_[i]), type(delim), i, dict_[i])
            s += "$%s%s%s%s" % (i, delim, dict_[i], delim)
        return s

    @classmethod
    def cal_datapos(cls, textend, datasize):
        start = 0
        end = 0
        start_new = textend + 1
        end_new = start_new + datasize - 1

        while True:
            inc_start = len(str(start_new)) - len(str(start))
            inc_end = len(str(end_new)) - len(str(end))
            start = start_new
            end = end_new

            if inc_start + inc_end > 0:
                start_new = inc_start + inc_end + start
                end_new = start_new + datasize - 1
            else:
                break

        return start, end


class DataSegment:
    def __init__(
        self, values: np.ndarray, datatype, num_dims, num_rows, endian
    ) -> None:
        self.values = values
        self.datatype = datatype
        self.num_dims = num_dims
        self.num_rows = num_rows
        self.endian = endian

    def to_string(self):
        format_ = self.endian + self.datatype.lower()
        return self.values.astype(format_).tobytes()

    @classmethod
    def from_string(cls, s: str, datatype, num_dims, num_rows, endian):
        data = BytesIO(s)
        format_ = endian + str(num_dims) + datatype.lower()
        datasize = struct.calcsize(format_)
        events = []
        for e in range(num_rows):
            read_data = data.read(datasize)
            event = struct.unpack(format_, read_data)
            events.append(event)
        values = np.array(events, dtype=np.float32 if datatype == "F" else np.float64)
        return DataSegment(values, datatype, num_dims, num_rows, endian)


class Fcs:
    def __init__(
        self,
        values: np.ndarray,
        short_channels: list,
        long_channels: list = None,
        ver: str = "FCS3.1",
        mode: str = "L",
        byte_order: str = "1,2,3,4",
        text_start: int = 256,
        datatype: str = "F",
    ) -> None:
        self.values = values
        self.short_channels = short_channels
        if long_channels is not None:
            self.long_channels = long_channels
        else:
            self.long_channels = [" " for _ in range(len(short_channels))]
        self.ver = ver
        self.mode = mode
        self.byte_order = byte_order

        self.tseg = TextSegment(
            values.shape[0],
            self.short_channels,
            self.long_channels,
            values.max(axis=0),
            mode=mode,
            byte_order=byte_order,
            text_start=text_start,
            datatype=datatype,
        )

        self.hseg = HeaderSegment(
            self.tseg.text_start,
            self.tseg.text_end,
            self.tseg.data_start,
            self.tseg.data_end,
            ver=self.ver,
        )

        self.dseg = DataSegment(
            values, datatype, values.shape[1], values.shape[0], self.tseg.endian
        )

    @classmethod
    @bufferize
    def from_file(cls, filepath_or_buffer):
        header = HeaderSegment.from_string(filepath_or_buffer.read(58))
        filepath_or_buffer.seek(header.text_start)
        fseg = TextSegment.from_string(
            filepath_or_buffer.read(header.text_end - header.text_start + 1)
        )
        data_start = header.data_start if header.data_start != 0 else fseg.data_start
        data_end = header.data_end if header.data_end != 0 else fseg.data_end
        filepath_or_buffer.seek(data_start)
        data = DataSegment.from_string(
            filepath_or_buffer.read(data_end - data_start + 1),
            fseg.datatype,
            len(fseg.pnn),
            fseg.tot,
            fseg.endian,
        )

        return Fcs(data.values, fseg.pnn, fseg.pns)

    @classmethod
    @bufferize
    def read_data_segment(cls, filepath_or_buffer, hseg, tseg):
        data_start = hseg.data_start if hseg.data_start != 0 else tseg.data_start
        data_end = hseg.data_end if hseg.data_end != 0 else tseg.data_end
        filepath_or_buffer.seek(data_start)
        return DataSegment.from_string(
            filepath_or_buffer.read(data_end - data_start + 1),
            tseg.datatype,
            len(tseg.pnn),
            tseg.tot,
            tseg.endian,
        )

    @classmethod
    @bufferize
    def read_header_segment(cls, filepath_or_buffer):
        filepath_or_buffer.seek(0)
        return HeaderSegment.from_string(filepath_or_buffer.read(58))

    @classmethod
    @bufferize
    def read_text_segment(cls, filepath_or_buffer):
        header = HeaderSegment.from_string(filepath_or_buffer.read(58))
        filepath_or_buffer.seek(header.text_start)
        fseg = TextSegment.from_string(
            filepath_or_buffer.read(header.text_end - header.text_start + 1)
        )

        return fseg

    @classmethod
    def write_bytes(cls, f, s):
        if not type(s) == bytes:
            f.write(s.encode("UTF-8"))
        else:
            f.write(s)

    @classmethod
    @bufferize(mode="rb+")
    def write_text_segment(cls, filepath_or_buffer, text_segment):
        filepath_or_buffer.seek(58)
        cls.write_bytes(
            filepath_or_buffer,
            " " * (text_segment.text_start - filepath_or_buffer.tell()),
        )
        cls.write_bytes(filepath_or_buffer, text_segment.to_string())
        cls.write_bytes(
            filepath_or_buffer,
            " " * (text_segment.data_start - filepath_or_buffer.tell()),
        )

    @classmethod
    @bufferize(mode="rb+")
    def write_header_segment(cls, filepath_or_buffer, header_segment):
        filepath_or_buffer.seek(0)
        cls.write_bytes(filepath_or_buffer, header_segment.to_string())

    @bufferize(mode="wb")
    def export(self, filepath_or_buffer):
        self.write_bytes(filepath_or_buffer, self.hseg.to_string())
        self.write_text_segment(filepath_or_buffer, self.tseg)
        self.write_bytes(filepath_or_buffer, self.dseg.to_string())

    @property
    def count(self):
        return self.tseg.tot



import re
from io import BytesIO
from abc import abstractmethod, ABC
from typing import Type
from contextlib import contextmanager
from functools import update_wrapper, partial
from ._typing import ReadFcsBuffer as ReadFcsBufferType
from ._typing import WriteFcsBuffer as WriteFcsBufferType
from typing import Union

try:
    import boto3
except ImportError:
    boto3 = None


class ReadFcsBuffer(ABC):
    @abstractmethod
    def seek(self, position: int):
        pass

    @abstractmethod
    def read(self, number: int):
        pass

    @abstractmethod
    def close(self):
        pass


class WriteFcsBuffer(ABC):
    @abstractmethod
    def seek(self, position: int):
        pass

    @abstractmethod
    def write(self, __b: str):
        pass

    @abstractmethod
    def tell(self):
        pass

    @abstractmethod
    def close(self):
        pass


class S3ReadBuffer(ReadFcsBuffer):
    def __init__(self, path, bucket):
        if boto3 is None:
            raise ImportError(
                "This module requires boto3. Please use 'pip install fcsy[boto3] or install boto3 manually"
            )
        super().__init__()
        self.path = path
        self.bucket = bucket
        self.position = 0
        self.s3 = boto3.client("s3")

    def seek(self, position: int):
        self.position = position

    def read(self, number: int):
        resp = self.s3.get_object(
            Bucket=self.bucket,
            Key=self.path,
            Range=f"bytes={self.position}-{self.position+number}",
        )
        self.position += number
        return resp["Body"].read(number)

    def close(self):
        pass


class S3WriteBuffer(WriteFcsBuffer):
    def __init__(self, path, bucket):
        if boto3 is None:
            raise ImportError(
                "This module requires boto3. Please use 'pip install fcsy[boto3] or install boto3 manually"
            )
        super().__init__()
        self.path = path
        self.bucket = bucket
        self.s3 = boto3.client("s3")
        self.buffer = BytesIO()

    def seek(self, position: int):
        self.buffer.seek(position)

    def tell(self):
        return self.buffer.tell()

    def write(self, bytes):
        self.buffer.write(bytes)

    def close(self):
        self.seek(0)
        self.s3.upload_fileobj(self.buffer, self.bucket, self.path)


def read_path(path):
    if path.startswith("s3://"):
        match = re.match(r"s3://(.*?)/(.*?)$", path)
        return {
            "mode": "s3",
            "contents": {"bucket": match.group(1), "key": match.group(2)},
        }
    else:
        return {"mode": "local", "contents": {"path": path}}


def create_open_func(
    parser_class: Union[Type[ReadFcsBufferType], Type[WriteFcsBufferType]],
    *args,
    **kwargs,
):
    @contextmanager
    def func(path, *func_args, **func_kwargs):
        obj = parser_class(path, *args, **kwargs)
        try:
            yield obj
        finally:
            obj.close()

    return func


class Bufferize:
    def __init__(self, func, mode="rb") -> None:
        update_wrapper(self, func)
        self.func = func
        self.mode = mode

    def __get__(self, obj, objtype):
        return partial(self.__call__, obj)

    def __call__(self, obj, filepath_or_buffer, *args, **kwargs):
        if type(filepath_or_buffer) is str:
            path = read_path(filepath_or_buffer)
            if path["mode"] == "s3":
                if self.mode == "rb":
                    buffer_class = S3ReadBuffer
                elif self.mode == "wb":
                    buffer_class = S3WriteBuffer
                else:
                    raise ValueError("invalid s3 buffer mode")

                open_func = create_open_func(
                    buffer_class, bucket=path["contents"]["bucket"]
                )
                with open_func(path["contents"]["key"], self.mode) as fp:
                    return self.func(obj, fp, *args, **kwargs)
            else:
                with open(filepath_or_buffer, self.mode) as fp:
                    return self.func(obj, fp, *args, **kwargs)
        else:
            return self.func(obj, filepath_or_buffer, *args, **kwargs)


def bufferize(func=None, mode="rb"):
    if func:
        return Bufferize(func)
    else:

        def wrapper(func):
            return Bufferize(func, mode=mode)

        return wrapper



__version__ = "0.10.0"

import pandas as pd
from copy import deepcopy
import warnings
from typing import Union
from .fcs import Fcs
from ._typing import ReadFcsBuffer, WriteFcsBuffer, UpdateFcsBuffer


__all__ = [
    "DataFrame",
    "Fcs",
    "write_fcs",
    "read_fcs",
    "read_fcs_names",
    "read_channels",
    "read_events_num",
    "rename_channels",
]


def read_fcs_names(f, name_type="short"):
    warnings.warn(
        "read_fcs_names() is deprecated; use read_channels()",
        DeprecationWarning,
    )
    fcs = Fcs.from_file(f)
    mapping = {"short": fcs.short_channels, "long": fcs.long_channels}
    return mapping[name_type]


def read_fcs(f, name_type="short"):
    warnings.warn(
        "read_fcs() is deprecated; use DataFrame.from_fcs()",
        DeprecationWarning,
    )
    return DataFrame.from_fcs(f, channel_type=name_type)


def write_fcs(df: pd.DataFrame, path: str, long_names=None):
    warnings.warn(
        "write_fcs() is deprecated; use DataFrame.export()",
        DeprecationWarning,
    )
    fcs = Fcs(df.values, df.columns, long_names)
    fcs.export(path)


class DataFrame(pd.DataFrame):
    @property
    def _constructor(self):
        return DataFrame

    def to_fcs(self, filepath_or_buffer: Union[str, WriteFcsBuffer]):

        """
        write fcs.

        :param filepath_or_buffer:
            String, or file-like object implementing a ``write()`` function.
            The string could be a s3 url with the format:
            ``s3://{bucket}/{key}``.
        :type param filepath_or_buffer: str or file-like object
        """

        long_channels = None
        if isinstance(self.columns, pd.MultiIndex):
            short_channels = self.columns.get_level_values("short").values
            long_channels = self.columns.get_level_values("long").values
        else:
            short_channels = self.columns

        Fcs(self.values, short_channels, long_channels).export(filepath_or_buffer)

    @classmethod
    def from_fcs(
        cls, filepath_or_buffer: Union[str, ReadFcsBuffer], channel_type: str = "short"
    ):
        """
        Read dataframe from fcs.

        :param filepath_or_buffer: str or file-like object.
            String, or file-like object implementing a ``read()`` function.
            The string could be a s3 url with the format:
            ``s3://{bucket}/{key}``.
        :type filepath_or_buffer: str or file-like object
        :param channel_type: {"short", "long", "multi"}, defaults to "short".
            "short" and "long" refer to short ($PnN) and long ($PnS) name of parameter n, respectively.
            See FCS3.1 data standard for detailed explanation.
        :type channel_type: str, optional
        :return: the dataframe contains the fcs channels and data
        :rtype: DataFrame
        """

        fcs = Fcs.from_file(filepath_or_buffer)
        colmap = {
            "short": fcs.short_channels,
            "long": fcs.long_channels,
            "multi": pd.MultiIndex.from_tuples(
                list(zip(fcs.short_channels, fcs.long_channels)),
                names=["short", "long"],
            ),
        }
        return DataFrame(fcs.values, columns=colmap[channel_type])


def read_channels(
    filepath_or_buffer: Union[str, ReadFcsBuffer], channel_type: str = "short"
) -> Union[list, pd.MultiIndex]:
    """
    Read fcs channels.

    :param filepath_or_buffer: String, or file-like object implementing a ``read()`` function.
            The string could be a s3 url with the format:
            ``s3://{bucket}/{key}``.
    :type filepath_or_buffer: str or file-like object
    :param channel_type: {"short", "long", "multi"}, defaults to "short".
        "short" and "long" refer to short ($PnN) and long ($PnS) name of parameter n, respectively.
        See FCS3.1 data standard for detailed explanation.
    :type channel_type: str, optional
    :return: list of the channels
    :rtype: Union[list, pd.MultiIndex]
    """

    fseg = Fcs.read_text_segment(filepath_or_buffer)
    maps = {
        "short": fseg.pnn,
        "long": fseg.pns,
        "multi": pd.MultiIndex.from_tuples(
            list(zip(fseg.pnn, fseg.pns)),
            names=["short", "long"],
        ),
    }
    return maps[channel_type]


def rename_channels(
    filepath_or_buffer: Union[str, UpdateFcsBuffer],
    channels: dict,
    channel_type: str,
    allow_rewrite: bool = False,
) -> None:
    """
    Rename fcs channels.

    :param filepath_or_buffer: String, or file-like object implementing both ``read()`` and ``write()`` functions.
            S3 url is not supported.
    :type filepath_or_buffer: str or file-like object
    :param channels: A mapper from old names to new names.
    :type channels: dict
    :param channel_type: {"short", "long"}.
        "short" and "long" refer to short ($PnN) and long ($PnS) name of parameter n, respectively.
        See FCS3.1 data standard for detailed explanation.
    :type channel_type: str
    :param allow_rewrite: Allow rewriting the whole file if channel only editing is not feasible.
        It can be one of the following:

        * The new channel names are too long causing Text Segment overlap with Data Segment.
        * filepath_or_buffer is s3 url.
    :type allow_rewrite: bool
    """
    tseg = Fcs.read_text_segment(filepath_or_buffer)
    tseg_old = deepcopy(tseg)
    {"short": tseg.update_pnn, "long": tseg.update_pns}[channel_type](channels)
    hseg = Fcs.read_header_segment(filepath_or_buffer)
    hseg.text_start = tseg.text_start
    hseg.text_end = tseg.text_end
    try:
        if tseg.text_end >= hseg.data_start:
            if allow_rewrite:
                dseg = Fcs.read_data_segment(filepath_or_buffer, hseg, tseg_old)
                Fcs(dseg.values, tseg.pnn, tseg.pns).export(filepath_or_buffer)
                return
            else:
                raise ValueError(
                    "New channel names are too long causing overlap with Data Segment."
                )
        Fcs.write_header_segment(filepath_or_buffer, hseg)
        Fcs.write_text_segment(filepath_or_buffer, tseg)
    except ValueError as e:
        if str(e) == "invalid s3 buffer mode":
            if allow_rewrite:
                dseg = Fcs.read_data_segment(filepath_or_buffer, hseg, tseg_old)
                Fcs(dseg.values, tseg.pnn, tseg.pns).export(filepath_or_buffer)
                return
            else:
                raise ValueError("S3 url is only supported when allow_rewrite is set.")
        else:
            raise e


def read_events_num(filepath_or_buffer: Union[str, ReadFcsBuffer]) -> int:
    """
    Read fcs events number.

    :param filepath_or_buffer: str or file-like object
            String, or file-like object implementing a ``read()`` function.
            The string could be a s3 url with the format:
            ``s3://{bucket}/{key}``.
    :type path: str
    :return: the events number
    :rtype: int
    """

    fseg = Fcs.read_text_segment(filepath_or_buffer)
    return fseg.tot



import os
import pytest

boto3 = pytest.importorskip("boto3")

from tempfile import mkdtemp
from shutil import rmtree
import pandas as pd
import numpy as np
from moto import mock_s3
from fcsy.fcs import Fcs
from fcsy.buffer import create_open_func, S3ReadBuffer, read_path
from fcsy import DataFrame


class TmpDir:
    def __enter__(self):
        self._tmp_dir_path = mkdtemp()
        return self._tmp_dir_path

    def __exit__(self, type, value, traceback):
        rmtree(self._tmp_dir_path)


class TestWithS3:
    mock_s3 = mock_s3()
    bucket_name = "test-bucket"

    def setup_method(self):
        self.name = "test.fcs"
        self.data = np.array(
            [[1.1, 2.1, 3.1, 4.0011], [11.1, 12.1, 13.1, 14.0011]], dtype=np.float32
        )
        self.channels = ["a", "b", "c", "d"]
        self.long_channels = ["A", "B", "C", "D"]

        self.mock_s3.start()
        # you can use boto3.client('s3') if you prefer
        self.s3 = boto3.client("s3", region_name="us-east-1")
        self.s3.create_bucket(Bucket=self.bucket_name)

    def teardown_method(self):
        self.mock_s3.stop()

    def test_read_text_segment(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, "test.fcs")
            cols = pd.MultiIndex.from_tuples(
                list(zip(self.channels, self.long_channels)),
                names=["short", "long"],
            )
            df = DataFrame(self.data, columns=cols)
            df.to_fcs(filename)

            self.s3.upload_file(filename, self.bucket_name, "test.fcs")

            parse_func = create_open_func(S3ReadBuffer, bucket=self.bucket_name)

            with parse_func("test.fcs") as fp:
                seg = Fcs.read_text_segment(fp)
            assert seg.pnn == self.channels
            assert seg.pns == self.long_channels

    def test_from_fcs(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, "test.fcs")
            cols = pd.MultiIndex.from_tuples(
                list(zip(self.channels, self.long_channels)),
                names=["short", "long"],
            )
            df = DataFrame(self.data, columns=cols)
            df.to_fcs(filename)

            self.s3.upload_file(filename, self.bucket_name, "test.fcs")
            parse_func = create_open_func(S3ReadBuffer, bucket=self.bucket_name)

            with parse_func("test.fcs") as fp:
                fcs = Fcs.from_file(fp)
                assert fcs.short_channels == self.channels
                assert fcs.long_channels == self.long_channels
                assert fcs.count == len(self.data)
                assert np.array_equal(fcs.values, self.data)


def test_read_path():
    assert read_path("s3://test-bucket/abc/testfile.fcs") == {
        "mode": "s3",
        "contents": {"bucket": "test-bucket", "key": "abc/testfile.fcs"},
    }
    assert read_path("test_dir/file1.fcs") == {
        "mode": "local",
        "contents": {"path": "test_dir/file1.fcs"},
    }



import numpy as np
import pandas as pd


def zscore(df, axis=0, ddof=0):
    a = np.asanyarray(df)
    mns = a.mean(axis=axis)
    sstd = a.std(axis=axis, ddof=ddof)
    if axis and mns.ndim < a.ndim:
        f = a - np.expand_dims(mns, axis=axis)
        t = np.expand_dims(sstd, axis=axis)
        x = f / t
    else:
        x = (a - mns) / sstd
    return pd.DataFrame(x, columns=df.columns, index=df.index)


def arcsinh(df, factor=5):
    df = df / factor
    return df.apply(np.arcsinh)


def rm_outliers(df, threshold=0.05, **kwargs):
    df = df.copy()
    num = int(df.shape[0] * threshold)
    for c in df:
        if c in kwargs:
            num = int(df.shape[0] * kwargs[c])
        ind = np.argsort(df[c].values)
        df.loc[ind[:num], c] = df.loc[ind[num], c]
        df.loc[ind[-num:], c] = df.loc[ind[-num - 1], c]
    return df


def randomize(df, factor=0.5):
    return df.applymap(lambda x: np.random.normal(x, factor))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from tempfile import mkdtemp
from shutil import rmtree
import pandas as pd
import numpy as np
from fcsy import DataFrame
from fcsy.fcs import TextSegment, HeaderSegment, DataSegment, Fcs


class TmpDir:
    def __enter__(self):
        self._tmp_dir_path = mkdtemp()
        return self._tmp_dir_path

    def __exit__(self, type, value, traceback):
        rmtree(self._tmp_dir_path)


class TestFCS:
    def setup_method(self):
        self.name = "test.fcs"
        self.data = np.array(
            [[1.1, 2.1, 3.1, 4.0011], [11.1, 12.1, 13.1, 14.0011]], dtype=np.float32
        )
        self.channels = ["a", "b", "c", "d"]
        self.long_channels = ["A", "B", "C", "D"]

    def test_read_text_segment(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, self.name)
            cols = pd.MultiIndex.from_tuples(
                list(zip(self.channels, self.long_channels)),
                names=["short", "long"],
            )
            df = DataFrame(self.data, columns=cols)
            df.to_fcs(filename)
            seg = Fcs.read_text_segment(filename)
            assert seg.pnn == self.channels
            assert seg.pns == self.long_channels

    def test_header(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, self.name)
            df = DataFrame(self.data, columns=self.channels)
            df.to_fcs(filename)
            fcs = Fcs.from_file(filename)
            assert fcs.short_channels == self.channels

    def test_count(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, self.name)
            df = DataFrame(self.data, columns=self.channels)
            df.to_fcs(filename)
            fcs = Fcs.from_file(filename)
            assert fcs.count == len(self.data)

    def test_data(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, self.name)
            df = DataFrame(self.data, columns=self.channels)
            df.to_fcs(filename)
            fcs = Fcs.from_file(filename)
            assert np.array_equal(fcs.values, self.data)

    def test_export(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, "export.fcs")
            text_start = 256
            fseg = TextSegment(
                self.data.shape[0],
                self.long_channels,
                self.channels,
                self.data.max(axis=0),
                text_start=text_start,
            )
            header = HeaderSegment(
                text_start, fseg.text_end, fseg.data_start, fseg.data_end
            )
            fcs = Fcs(self.data, self.long_channels, self.channels)
            fcs.export(filename)

            with open(filename, "rb") as fp:
                assert fp.read(58).decode("UTF-8") == header.to_string()
                fp.seek(text_start)
                assert (
                    fp.read(header.text_end - fseg.text_start + 1)
                    .decode("UTF-8")
                    .strip()
                    == fseg.to_string()
                )
                fp.seek(fseg.data_start)
                dseg = DataSegment.from_string(
                    fp.read(fseg.data_end - fseg.data_start + 1),
                    fseg.datatype,
                    len(fseg.pnn),
                    fseg.tot,
                    fseg.endian,
                )
                assert np.array_equal(dseg.values, self.data)

    def test_buffer_export(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, "export.fcs")
            text_start = 256
            fseg = TextSegment(
                self.data.shape[0],
                self.long_channels,
                self.channels,
                self.data.max(axis=0),
                text_start=text_start,
            )
            header = HeaderSegment(
                text_start, fseg.text_end, fseg.data_start, fseg.data_end
            )
            fcs = Fcs(self.data, self.long_channels, self.channels)
            with open(filename, "wb") as fp:
                fcs.export(fp)

            with open(filename, "rb") as fp:
                assert fp.read(58).decode("UTF-8") == header.to_string()
                fp.seek(text_start)
                assert (
                    fp.read(header.text_end - fseg.text_start + 1)
                    .decode("UTF-8")
                    .strip()
                    == fseg.to_string()
                )
                fp.seek(fseg.data_start)
                dseg = DataSegment.from_string(
                    fp.read(fseg.data_end - fseg.data_start + 1),
                    fseg.datatype,
                    len(fseg.pnn),
                    fseg.tot,
                    fseg.endian,
                )
                assert np.array_equal(dseg.values, self.data)

    def test_write_text_segment(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, "export.fcs")
            fcs = Fcs(self.data, self.long_channels, self.channels)
            fcs.export(filename)

            fseg = TextSegment(
                self.data.shape[0],
                self.long_channels,
                ["w", "x", "y", "z"],
                self.data.max(axis=0),
                text_start=256,
            )

            Fcs.write_text_segment(filename, fseg)
            tseg = Fcs.read_text_segment(filename)
            assert tseg.pns == ["w", "x", "y", "z"]


class TestTextSegment:
    def setup_method(self):
        self.name = "test.fcs"
        self.data = np.array(
            [[1.1, 2.1, 3.1, 4.0011], [11.1, 12.1, 13.1, 14.0011]], dtype=np.float32
        )
        self.channels = ["a", "b", "c", "d"]
        self.long_channels = ["A", "B", "C", "D"]

    def test_locations(self):
        fseg = TextSegment(
            self.data.shape[0],
            self.long_channels,
            self.channels,
            self.data.max(axis=0),
            data_start=1000,
        )
        assert fseg.text_start == 256
        assert fseg.text_end == 564
        assert fseg.data_start == 1000
        assert fseg.data_end == 0

    def test_data_pos(self):
        assert TextSegment.cal_datapos(10, 100) == (14, 113)

    def test_dict_to_string(self):
        text = {}
        text["BEGINDATA"] = "256"
        text["BYTEORD"] = "1,2,3,4"  # little endian
        assert (
            TextSegment.dict_to_string(text, "/") == "/$BEGINDATA/256/$BYTEORD/1,2,3,4/"
        )

    def test_from_string(self):
        with open("tests/TextSegment.txt", "rb") as fp:
            s = fp.readline()

        fseq = TextSegment.from_string(s)
        assert fseq.data_start == 4064
        assert fseq.data_end == 119207043

    def test_cal_max_text_ends(self):
        assert TextSegment.cal_max_text_ends({}, "/", 0) == 124

    def test_cal_datapos(self):
        textend = 2
        datasize = 8
        start_expected = 4
        end_expected = 11
        start, end = TextSegment.cal_datapos(textend, datasize)
        assert start == start_expected
        assert end == end_expected

        textend = 999
        datasize = 9000
        start_expected = 1007
        end_expected = 10006
        start, end = TextSegment.cal_datapos(textend, datasize)
        assert start == start_expected
        assert end == end_expected

    def test_update_pns(self):
        fseg = TextSegment(
            self.data.shape[0],
            self.long_channels,
            self.channels,
            self.data.max(axis=0),
            data_start=1000,
        )
        fseg.update_pns({"a": "aa", "b": "bb"})
        assert fseg.pns == ["aa", "bb", "c", "d"]

        text_str = fseg.to_string()
        assert "$P1S/aa" in text_str
        assert "$P2S/bb" in text_str

    def test_update_pnn(self):
        fseg = TextSegment(
            self.data.shape[0],
            self.long_channels,
            self.channels,
            self.data.max(axis=0),
            data_start=1000,
        )
        fseg.update_pnn({"A": "AA", "B": "BB"})
        assert fseg.pnn == ["AA", "BB", "C", "D"]

        text_str = fseg.to_string()
        assert "$P1N/AA" in text_str
        assert "$P2N/BB" in text_str


class TestHeaderSegment:
    def setup_method(self):
        self.name = "test.fcs"
        self.data = np.array(
            [[1.1, 2.1, 3.1, 4.0011], [11.1, 12.1, 13.1, 14.0011]], dtype=np.float32
        )
        self.channels = ["a", "b", "c", "d"]
        self.long_channels = ["A", "B", "C", "D"]

    def test_to_string(self):
        header = HeaderSegment(256, 561, 563, 99999999, 0, 0)
        assert (
            header.to_string()
            == "FCS3.1         256     561     56399999999       0       0"
        )

        header = HeaderSegment(256, 561, 563, 100000000, 0, 0)
        assert (
            header.to_string()
            == "FCS3.1         256     561       0       0       0       0"
        )

    def test_from_string(self):
        s = b"FCS3.1         256     561     563     789    1000    2000"
        header = HeaderSegment.from_string(s)
        assert header.asdict() == dict(
            text_start=256,
            text_end=561,
            data_start=563,
            data_end=789,
            analysis_start=1000,
            analysis_end=2000,
        )
        assert header.ver == "FCS3.1"


class TestDataSegment:
    def setup_method(self):
        self.name = "test.fcs"
        self.data = np.array(
            [[1.1, 2.1, 3.1, 4.0011], [11.1, 12.1, 13.1, 14.0011]], dtype=np.float32
        )
        self.channels = ["a", "b", "c", "d"]
        self.long_channels = ["A", "B", "C", "D"]

    def test_to_string(self):
        data = DataSegment(self.data, "f", self.data.shape[1], self.data.shape[0], "<")
        assert (
            data.to_string()
            == b"\xcd\xcc\x8c?ff\x06@ffF@\x03\t\x80@\x9a\x991A\x9a\x99AA\x9a\x99QA\x81\x04`A"
        )

    def test_from_string(self):
        s = b"\xcd\xcc\x8c?ff\x06@ffF@\x03\t\x80@\x9a\x991A\x9a\x99AA\x9a\x99QA\x81\x04`A"
        data = DataSegment.from_string(
            s, "f", self.data.shape[1], self.data.shape[0], "<"
        )
        np.testing.assert_array_equal(data.values, self.data)



import os
from tempfile import mkdtemp
from shutil import rmtree
import numpy as np
from numpy.testing import assert_almost_equal
import pandas as pd
from pandas.testing import assert_frame_equal
import pytest
from fcsy import (
    DataFrame,
    read_channels,
    read_events_num,
    read_fcs,
    write_fcs,
    rename_channels,
)


class TmpDir:
    def __enter__(self):
        self._tmp_dir_path = mkdtemp()
        return self._tmp_dir_path

    def __exit__(self, type, value, traceback):
        rmtree(self._tmp_dir_path)


class TestDataFrame:
    def setup_method(self):
        self.name = "test.fcs"
        self.data = np.array(
            [[1.1, 2.1, 3.1, 4.0011], [11.1, 12.1, 13.1, 14.0011]], dtype=np.float32
        )
        self.short_channels = ["a_di", "b_di", "c_di", "d_di"]
        self.long_channels = ["A A", "B B", "C C", "D D"]

    def test_io(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, self.name)
            df = DataFrame(
                self.data,
                columns=pd.MultiIndex.from_tuples(
                    zip(self.short_channels, self.long_channels),
                    names=["short", "long"],
                ),
            )
            df.to_fcs(filename)
            df = DataFrame.from_fcs(filename, "multi")
            np.testing.assert_array_equal(df.values, self.data)
            np.testing.assert_array_equal(
                df.columns,
                pd.MultiIndex.from_tuples(
                    zip(self.short_channels, self.long_channels),
                    names=["short", "long"],
                ),
            )

    def test_read_fcs(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, self.name)
            df = DataFrame(self.data, columns=self.short_channels)
            df.to_fcs(filename)
            df2 = read_fcs(filename)
            assert_frame_equal(df, df2)

    def test_write_fcs(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, self.name)
            df = DataFrame(self.data, columns=self.short_channels)
            write_fcs(df, filename, long_names=self.long_channels)
            df = DataFrame.from_fcs(filename, "multi")
            np.testing.assert_array_equal(df.values, self.data)
            np.testing.assert_array_equal(
                df.columns,
                pd.MultiIndex.from_tuples(
                    zip(self.short_channels, self.long_channels),
                    names=["short", "long"],
                ),
            )

    def test_read_events_num(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, self.name)
            df = DataFrame(self.data, columns=self.short_channels)
            write_fcs(df, filename, long_names=self.long_channels)
            num = read_events_num(filename)
            assert num == 2

    def test_read_channels(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, self.name)
            df = DataFrame(self.data, columns=self.short_channels)
            write_fcs(df, filename, long_names=self.long_channels)

            assert read_channels(filename, "short") == self.short_channels
            assert read_channels(filename, "long") == self.long_channels
            np.testing.assert_array_equal(
                read_channels(filename, "multi"),
                pd.MultiIndex.from_tuples(
                    zip(self.short_channels, self.long_channels),
                    names=["short", "long"],
                ),
            )

    def test_rename_channels(self):
        with TmpDir() as dir_:
            df = DataFrame(
                self.data,
                columns=pd.MultiIndex.from_tuples(
                    zip(self.short_channels, self.long_channels),
                    names=["short", "long"],
                ),
            )
            filename = os.path.join(dir_, self.name)
            df.to_fcs(filename)
            rename_channels(
                filename, dict(zip(self.short_channels, list("abcd"))), "short"
            )
            short_channels = read_channels(filename, "short")
            assert short_channels == list("abcd")

            rename_channels(
                filename, dict(zip(self.long_channels, list("ABCD"))), "long"
            )
            long_channels = read_channels(filename, "long")
            assert long_channels == list("ABCD")

            data = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
            columns = pd.MultiIndex.from_tuples(
                list(zip("abc", "ABC")), names=["short", "long"]
            )
            df = DataFrame(data, columns=columns)
            filename = os.path.join(dir_, self.name)
            df.to_fcs(filename)
            rename_channels(
                filename, {"a": "a_1", "b": "b_1", "c": "c_1"}, channel_type="short"
            )
            short_channels = read_channels(filename, "short")
            long_channels = read_channels(filename, "long")
            assert short_channels == ["a_1", "b_1", "c_1"]
            assert long_channels == list("ABC")

    def test_rename_channels_alt(self):
        with TmpDir() as dir_:
            df = DataFrame(
                self.data,
                columns=pd.MultiIndex.from_tuples(
                    zip(self.short_channels, self.long_channels),
                    names=["short", "long"],
                ),
            )
            filename = os.path.join(dir_, self.name)
            df.to_fcs(filename)

            ln = ""
            for i in range(30):
                ln = ln + str(i)

            with pytest.raises(ValueError) as err:
                rename_channels(filename, {"a_di": ln}, "short")
            assert (
                str(err.value)
                == "New channel names are too long causing overlap with Data Segment."
            )

            rename_channels(filename, {"a_di": ln}, "short", allow_rewrite=True)
            df2 = DataFrame.from_fcs(filename)

            assert_almost_equal(df.values, df2.values)


class TestWithBuffer:
    def setup_method(self):
        self.name = "test.fcs"
        self.data = np.array(
            [[1.1, 2.1, 3.1, 4.0011], [11.1, 12.1, 13.1, 14.0011]], dtype=np.float32
        )
        self.short_channels = ["a", "b", "c", "d"]
        self.long_channels = ["A", "B", "C", "D"]

    def test_from_fcs(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, "test.fcs")
            cols = pd.MultiIndex.from_tuples(
                list(zip(self.short_channels, self.long_channels)),
                names=["short", "long"],
            )
            df = DataFrame(self.data, columns=cols)
            df.to_fcs(filename)

            with open(filename, "rb") as fp:
                dfr = DataFrame.from_fcs(fp, channel_type="multi")
            assert_frame_equal(dfr, df)

    def test_to_fcs(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, "test.fcs")
            cols = pd.MultiIndex.from_tuples(
                list(zip(self.short_channels, self.long_channels)),
                names=["short", "long"],
            )
            df = DataFrame(self.data, columns=cols)
            with open(filename, "wb") as fp:
                df.to_fcs(fp)

            dfr = DataFrame.from_fcs(filename, channel_type="multi")
            assert_frame_equal(df, dfr)

    def test_read_events_num(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, "test.fcs")
            cols = pd.MultiIndex.from_tuples(
                list(zip(self.short_channels, self.long_channels)),
                names=["short", "long"],
            )
            df = DataFrame(self.data, columns=cols)
            df.to_fcs(filename)

            with open(filename, "rb") as fp:
                num = read_events_num(fp)
            assert num == len(self.data)

    def test_read_channels(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, "test.fcs")
            cols = pd.MultiIndex.from_tuples(
                list(zip(self.short_channels, self.long_channels)),
                names=["short", "long"],
            )
            df = DataFrame(self.data, columns=cols)
            df.to_fcs(filename)

            assert read_channels(filename, "short") == self.short_channels
            assert read_channels(filename, "long") == self.long_channels
            np.testing.assert_array_equal(read_channels(filename, "multi"), cols)

    def test_rename_channels(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, "test.fcs")
            cols = pd.MultiIndex.from_tuples(
                list(zip(self.short_channels, self.long_channels)),
                names=["short", "long"],
            )
            df = DataFrame(self.data, columns=cols)
            df.to_fcs(filename)

            with open(filename, "rb+") as fp:
                rename_channels(
                    fp, dict(zip(self.short_channels, list("wxyz"))), "short"
                )

            with open(filename, "rb+") as fp:
                rename_channels(fp, dict(zip(self.long_channels, list("WXYZ"))), "long")

            assert read_channels(filename, "short") == list("wxyz")
            assert read_channels(filename, "long") == list("WXYZ")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import os
from setuptools import setup, find_packages


with open("README.md") as readme_file:
    readme = readme_file.read()

# with open("HISTORY.rst") as history_file:
#     history = history_file.read()

with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest",
]

extras_require = {"s3": ["boto3>=1.10.0"]}

setup(
    author="yc",
    author_email="yang.chen@scilifelab.se",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="A package for processing FCS files",
    install_requires=requirements,
    license="MIT license",
    long_description_content_type="text/markdown",
    long_description=readme,
    include_package_data=True,
    keywords="fcsy",
    name="fcsy",
    packages=find_packages(include=["fcsy"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/nehcgnay/fcsy",
    version="0.10.0",
    zip_safe=False,
    extras_require=extras_require,
)



from io import BufferedRandom
from typing import Any

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol


class ReadFcsBuffer(Protocol):
    def seek(self, position: int):
        pass

    def read(self, number: int):
        pass

    def close(self) -> Any:
        pass


class WriteFcsBuffer(Protocol):
    def seek(self, position: int):
        pass

    def tell(self) -> int:
        pass

    def write(self, __b: str) -> Any:
        pass

    def close(self) -> Any:
        pass


UpdateFcsBuffer = BufferedRandom



import os
import pytest

boto3 = pytest.importorskip("boto3")
from moto import mock_s3
from tempfile import mkdtemp
from shutil import rmtree
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from fcsy import (
    DataFrame,
    read_channels,
    read_events_num,
    rename_channels,
)


class TmpDir:
    def __enter__(self):
        self._tmp_dir_path = mkdtemp()
        return self._tmp_dir_path

    def __exit__(self, type, value, traceback):
        rmtree(self._tmp_dir_path)


class TestWithS3:
    mock_s3 = mock_s3()
    bucket_name = "test-bucket"

    def setup_method(self):
        self.name = "test.fcs"
        self.data = np.array(
            [[1.1, 2.1, 3.1, 4.0011], [11.1, 12.1, 13.1, 14.0011]], dtype=np.float32
        )
        self.short_channels = ["a", "b", "c", "d"]
        self.long_channels = ["A", "B", "C", "D"]

        self.mock_s3.start()
        self.s3 = boto3.client("s3", region_name="us-east-1")
        self.s3.create_bucket(Bucket=self.bucket_name)

    def teardown_method(self):
        self.mock_s3.stop()

    def test_from_fcs(self):
        with TmpDir() as dir_:
            filename = os.path.join(dir_, "test.fcs")
            cols = pd.MultiIndex.from_tuples(
                list(zip(self.short_channels, self.long_channels)),
                names=["short", "long"],
            )
            df = DataFrame(self.data, columns=cols)
            df.to_fcs(filename)

            self.s3.upload_file(filename, self.bucket_name, "test.fcs")

            dfr = DataFrame.from_fcs(
                f"s3://{self.bucket_name}/test.fcs", channel_type="multi"
            )
            assert_frame_equal(dfr, df)

    def test_to_fcs(self):
        with TmpDir() as dir_:
            cols = pd.MultiIndex.from_tuples(
                list(zip(self.short_channels, self.long_channels)),
                names=["short", "long"],
            )
            df = DataFrame(self.data, columns=cols)
            df.to_fcs(f"s3://{self.bucket_name}/test.fcs")
            filename = os.path.join(dir_, "test.fcs")
            self.s3.download_file(self.bucket_name, "test.fcs", filename)

            dfr = DataFrame.from_fcs(filename, channel_type="multi")
            assert_frame_equal(df, dfr)

    def test_read_events_num(self):
        df = DataFrame(self.data, columns=self.short_channels)
        s3_file_path = f"s3://{self.bucket_name}/test.fcs"
        df.to_fcs(s3_file_path)

        num = read_events_num(s3_file_path)
        assert num == len(self.data)

    def test_read_channels(self):
        cols = pd.MultiIndex.from_tuples(
            list(zip(self.short_channels, self.long_channels)),
            names=["short", "long"],
        )
        df = DataFrame(self.data, columns=cols)
        s3_file_path = f"s3://{self.bucket_name}/test.fcs"
        df.to_fcs(s3_file_path)
        assert read_channels(s3_file_path, "short") == self.short_channels
        assert read_channels(s3_file_path, "long") == self.long_channels
        np.testing.assert_array_equal(read_channels(s3_file_path, "multi"), cols)

    def test_rename_channels_alt(self):
        cols = pd.MultiIndex.from_tuples(
            list(zip(self.short_channels, self.long_channels)),
            names=["short", "long"],
        )
        df = DataFrame(self.data, columns=cols)
        s3_file_path = f"s3://{self.bucket_name}/test.fcs"
        df.to_fcs(s3_file_path)

        with pytest.raises(ValueError) as e:
            rename_channels(
                s3_file_path, dict(zip(self.short_channels, list("abcd"))), "short"
            )
        assert str(e.value) == "S3 url is only supported when allow_rewrite is set."

        ln = ""
        for i in range(30):
            ln = ln + str(i)
        with pytest.raises(ValueError) as e:
            rename_channels(s3_file_path, {"c": ln}, "short")
        assert (
            str(e.value)
            == "New channel names are too long causing overlap with Data Segment."
        )

    def test_rename_channels(self):
        cols = pd.MultiIndex.from_tuples(
            list(zip(self.short_channels, self.long_channels)),
            names=["short", "long"],
        )
        df = DataFrame(self.data, columns=cols)
        s3_file_path = f"s3://{self.bucket_name}/test.fcs"
        df.to_fcs(s3_file_path)

        rename_channels(
            s3_file_path, {"a": "a_1", "d": "d_1"}, "short", allow_rewrite=True
        )

        dfr = DataFrame.from_fcs(s3_file_path, channel_type="multi")

        np.testing.assert_array_equal(
            ["a_1", "b", "c", "d_1"],
            dfr.columns.get_level_values("short").values,
        )
        np.testing.assert_array_equal(
            list("ABCD"),
            dfr.columns.get_level_values("long").values,
        )

        ln = ""
        for i in range(30):
            ln = ln + str(i)
        rename_channels(s3_file_path, {"C": ln}, "long", allow_rewrite=True)
        assert ["A", "B", ln, "D"] == read_channels(s3_file_path, "long")
